# Copyright (c) 2026 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import os
import tempfile
import textwrap
import unittest
from contextlib import contextmanager
from pathlib import Path
from unittest.mock import patch

from setuptools import find_packages


def _setup_script_path():
    return Path(__file__).resolve().parents[1] / "build" / "scripts" / "setup"


def _write_pyproject(tmpdir, content):
    Path(tmpdir).joinpath("pyproject.toml").write_text(content, encoding="utf-8")


def _run_setup_script():
    path = _setup_script_path()
    code = path.read_text(encoding="utf-8")
    exec(compile(code, str(path), "exec"), {"__name__": "__main__"})


def _minimal_pyproject(
    extra="",
    license_line='license = "LGPL-3.0-or-later"',
    authors_block=textwrap.dedent(
        """\
        authors = [
            {name = "Test Author", email = "test@example.com"},
        ]
        """
    ),
):
    content = textwrap.dedent(
        f"""\
        [project]
        name = "test-pkg"
        version = "1.0.0"
        {license_line}
        """
    )
    if authors_block:
        content += textwrap.dedent(authors_block)
    if extra:
        content += textwrap.dedent(extra)
    return content


@contextmanager
def _project_dir(pyproject_content, layout=None):
    previous_cwd = os.getcwd()
    with tempfile.TemporaryDirectory() as tmpdir:
        _write_pyproject(tmpdir, pyproject_content)
        if layout:
            layout(Path(tmpdir))
        os.chdir(tmpdir)
        try:
            yield tmpdir
        finally:
            os.chdir(previous_cwd)


class TestSetupScript(unittest.TestCase):
    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_missing_project_section_exits(self, mock_setup, mock_exit):
        mock_exit.side_effect = SystemExit
        pyproject = textwrap.dedent(
            """\
            [build-system]
            requires = ['setuptools']
            """
        )
        with _project_dir(pyproject):
            with self.assertRaises(SystemExit):
                _run_setup_script()

        mock_exit.assert_called_once_with(0)
        mock_setup.assert_not_called()

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_minimal_project_metadata(self, mock_setup, mock_exit):
        with _project_dir(_minimal_pyproject()):
            _run_setup_script()

        mock_exit.assert_not_called()
        mock_setup.assert_called_once_with(
            name="test-pkg",
            version="1.0.0",
            author="Test Author",
            author_email="test@example.com",
            platforms=["GNU/Linux"],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_no_authors(self, mock_setup, mock_exit):
        with _project_dir(_minimal_pyproject(authors_block="")):
            _run_setup_script()

        mock_exit.assert_not_called()
        mock_setup.assert_called_once_with(
            name="test-pkg",
            version="1.0.0",
            platforms=["GNU/Linux"],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_author_name_only(self, mock_setup, mock_exit):
        authors_block = textwrap.dedent(
            """\
            authors = [
                {name = "Test Author"},
            ]
            """
        )
        with _project_dir(_minimal_pyproject(authors_block=authors_block)):
            _run_setup_script()

        mock_exit.assert_not_called()
        mock_setup.assert_called_once_with(
            name="test-pkg",
            version="1.0.0",
            author="Test Author",
            platforms=["GNU/Linux"],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_author_email_only(self, mock_setup, mock_exit):
        authors_block = textwrap.dedent(
            """\
            authors = [
                {email = "test@example.com"},
            ]
            """
        )
        with _project_dir(_minimal_pyproject(authors_block=authors_block)):
            _run_setup_script()

        mock_exit.assert_not_called()
        mock_setup.assert_called_once_with(
            name="test-pkg",
            version="1.0.0",
            author_email="test@example.com",
            platforms=["GNU/Linux"],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_console_scripts_entry_points(self, mock_setup, mock_exit):
        extra = """
        [project.scripts]
        my-cmd = "mypkg.cli:main"
        """
        with _project_dir(_minimal_pyproject(extra)):
            _run_setup_script()

        mock_setup.assert_called_once()
        kwargs = mock_setup.call_args[1]
        self.assertEqual(
            kwargs["entry_points"],
            {"console_scripts": ["my-cmd=mypkg.cli:main"]},
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_namespace_packages_explicit_list(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.packages.find]
        include = ["rfl.build*"]
        """

        def layout(root):
            rfl_build = root / "rfl" / "build"
            rfl_build.mkdir(parents=True)
            rfl_build.joinpath("__init__.py").write_text("", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(mock_setup.call_args[1]["packages"], ["rfl.build"])

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_autofind_packages(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.packages.find]
        include = ["mypkg.*"]
        """

        def layout(root):
            mypkg = root / "mypkg"
            mypkg.mkdir()
            mypkg.joinpath("__init__.py").write_text("", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()
            expected_packages = find_packages()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["packages"],
            expected_packages,
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_autofind_packages_with_top_level_wildcard(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.packages.find]
        include = ["mypkg*"]
        """

        def layout(root):
            mypkg = root / "mypkg"
            mypkg.mkdir()
            mypkg.joinpath("__init__.py").write_text("", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()
            expected_packages = find_packages()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["packages"],
            expected_packages,
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_autofind_flat_layout_ignores_include_filter(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.packages.find]
        include = ["mypkg.*"]
        """

        def layout(root):
            mypkg = root / "mypkg"
            mypkg.mkdir()
            mypkg.joinpath("__init__.py").write_text("", encoding="utf-8")
            other = root / "other"
            other.mkdir()
            other.joinpath("__init__.py").write_text("", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()
            expected_packages = find_packages()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["packages"],
            expected_packages,
        )
        self.assertEqual(sorted(expected_packages), ["mypkg", "other"])

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_autofind_packages_src_layout_where_only(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.packages.find]
        where = ["src"]
        """

        def layout(root):
            mypkg = root / "src" / "mypkg"
            mypkg.mkdir(parents=True)
            mypkg.joinpath("__init__.py").write_text("", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()
            expected_packages = find_packages(where="src")

        mock_setup.assert_called_once()
        kwargs = mock_setup.call_args[1]
        self.assertEqual(kwargs["packages"], expected_packages)
        self.assertEqual(kwargs["package_dir"], {"": "src"})

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_namespace_packages_with_where(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.packages.find]
        where = ["src"]
        include = ["myapp.core*"]
        """

        def layout(root):
            myapp_core = root / "src" / "myapp" / "core"
            myapp_core.mkdir(parents=True)
            myapp_core.joinpath("__init__.py").write_text("", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()

        mock_setup.assert_called_once()
        kwargs = mock_setup.call_args[1]
        self.assertEqual(kwargs["packages"], ["myapp.core"])
        self.assertEqual(kwargs["package_dir"], {"": "src"})

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_autofind_packages_src_layout_with_include(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.packages.find]
        where = ["src"]
        include = ["mypkg.*"]
        """

        def layout(root):
            mypkg = root / "src" / "mypkg"
            mypkg.mkdir(parents=True)
            mypkg.joinpath("__init__.py").write_text("", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()
            expected_packages = find_packages(
                where="src",
                include=("mypkg.*",),
                exclude=(),
            )

        mock_setup.assert_called_once()
        kwargs = mock_setup.call_args[1]
        self.assertEqual(kwargs["packages"], expected_packages)
        self.assertEqual(kwargs["package_dir"], {"": "src"})

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_explicit_packages_without_find(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools]
        packages = ["pkg.a", "pkg.b"]
        """
        with _project_dir(_minimal_pyproject(extra)):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["packages"],
            ["pkg.a", "pkg.b"],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_package_data(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.package-data]
        "pkg.data" = ["*.json"]
        """
        with _project_dir(_minimal_pyproject(extra)):
            _run_setup_script()

        mock_setup.assert_called_once()
        kwargs = mock_setup.call_args[1]
        self.assertEqual(kwargs["package_data"], {"pkg.data": ["*.json"]})
        self.assertTrue(kwargs["include_package_data"])

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_data_files_literals(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.data-files]
        "slurm-quota/conf" = [
            "conf/serve.yml",
            "conf/serve.ini.example",
        ]
        """

        def layout(root):
            conf = root / "conf"
            conf.mkdir()
            conf.joinpath("serve.yml").write_text("yml", encoding="utf-8")
            conf.joinpath("serve.ini.example").write_text("ini", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["data_files"],
            [
                (
                    "slurm-quota/conf",
                    ["conf/serve.yml", "conf/serve.ini.example"],
                ),
            ],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_data_files_glob_expansion(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.data-files]
        "slurm-quota/web/templates" = ["web/templates/*"]
        "slurm-quota/web/static" = ["web/static/*"]
        """

        def layout(root):
            templates = root / "web" / "templates"
            templates.mkdir(parents=True)
            templates.joinpath("a.html").write_text("html", encoding="utf-8")
            static = root / "web" / "static"
            static.mkdir(parents=True)
            static.joinpath("b.css").write_text("css", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["data_files"],
            [
                ("slurm-quota/web/templates", ["web/templates/a.html"]),
                ("slurm-quota/web/static", ["web/static/b.css"]),
            ],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_data_files_mixed_literals_and_globs(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.data-files]
        "slurm-quota/conf" = [
            "conf/serve.yml",
            "conf/serve.ini.example",
            "conf/slurm-quota-web.default",
        ]
        "slurm-quota/web/wsgi" = ["web/wsgi/slurm-quota-web.wsgi"]
        "slurm-quota/web/templates" = ["web/templates/*"]
        "slurm-quota/web/static" = ["web/static/*"]
        """

        def layout(root):
            conf = root / "conf"
            conf.mkdir()
            conf.joinpath("serve.yml").write_text("yml", encoding="utf-8")
            conf.joinpath("serve.ini.example").write_text("ini", encoding="utf-8")
            conf.joinpath("slurm-quota-web.default").write_text(
                "default", encoding="utf-8"
            )
            wsgi = root / "web" / "wsgi"
            wsgi.mkdir(parents=True)
            wsgi.joinpath("slurm-quota-web.wsgi").write_text("wsgi", encoding="utf-8")
            templates = root / "web" / "templates"
            templates.mkdir(parents=True)
            templates.joinpath("index.html").write_text("html", encoding="utf-8")
            static = root / "web" / "static"
            static.mkdir(parents=True)
            static.joinpath("app.css").write_text("css", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["data_files"],
            [
                (
                    "slurm-quota/conf",
                    [
                        "conf/serve.yml",
                        "conf/serve.ini.example",
                        "conf/slurm-quota-web.default",
                    ],
                ),
                (
                    "slurm-quota/web/wsgi",
                    ["web/wsgi/slurm-quota-web.wsgi"],
                ),
                ("slurm-quota/web/templates", ["web/templates/index.html"]),
                ("slurm-quota/web/static", ["web/static/app.css"]),
            ],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_install_requires(self, mock_setup, mock_exit):
        extra = """
        dependencies = [
            "dep-a",
            "dep-b",
        ]
        """
        with _project_dir(_minimal_pyproject(extra)):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["install_requires"],
            ["dep-a", "dep-b"],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_extras_require(self, mock_setup, mock_exit):
        extra = """
        [project.optional-dependencies]
        dev = ["pytest"]
        docs = ["sphinx"]
        """
        with _project_dir(_minimal_pyproject(extra)):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["extras_require"],
            {"dev": ["pytest"], "docs": ["sphinx"]},
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_license_text(self, mock_setup, mock_exit):
        with _project_dir(
            _minimal_pyproject(license_line='license = { text = "MIT" }')
        ):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(mock_setup.call_args[1]["license"], "MIT")

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_license_file(self, mock_setup, mock_exit):
        with _project_dir(
            _minimal_pyproject(license_line='license = { file = "LICENSE" }')
        ):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["license_files"],
            ["LICENSE"],
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_homepage_url(self, mock_setup, mock_exit):
        extra = """
        [project.urls]
        Homepage = "https://example.com"
        """
        with _project_dir(_minimal_pyproject(extra)):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(
            mock_setup.call_args[1]["url"],
            "https://example.com",
        )

    @patch("sys.exit")
    @patch("setuptools.setup")
    def test_skip_bare_include_pattern(self, mock_setup, mock_exit):
        extra = """
        [tool.setuptools.packages.find]
        include = ["rfl", "rfl.build*"]
        """

        def layout(root):
            rfl_build = root / "rfl" / "build"
            rfl_build.mkdir(parents=True)
            rfl_build.joinpath("__init__.py").write_text("", encoding="utf-8")

        with _project_dir(_minimal_pyproject(extra), layout=layout):
            _run_setup_script()

        mock_setup.assert_called_once()
        self.assertEqual(mock_setup.call_args[1]["packages"], ["rfl.build"])



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
from pathlib import Path

from rfl.build.ninja import NinjaBuilder


class TestNinjaBuilder(unittest.TestCase):
    def test_ninja_builder_basic(self):
        builder = NinjaBuilder()
        builder.variable("var1", "value1")
        builder.rule(name="build-cmd", command="build-stuff $in $out")
        builder.build(
            outputs=[Path("/output")], rule="build-cmd", inputs=[Path("/input")]
        )



# Copyright (c) 2025 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest

from rfl.authentication.user import AuthenticatedUser, AnonymousUser


class TestUser(unittest.TestCase):
    def test_user(self):
        user = AuthenticatedUser("test")
        self.assertFalse(user.is_anonymous())

    def test_anonymous(self):
        user = AnonymousUser()
        self.assertTrue(user.is_anonymous())



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import os
import unittest
from unittest.mock import patch, Mock
from pathlib import Path
import urllib
import ssl

import ldap
import ldap.filter

from rfl.authentication.ldap import LDAPAuthentifier
from rfl.authentication.errors import LDAPAuthenticationError


class MockLDAPObject:
    pass


class TestLDAPAuthentifier(unittest.TestCase):
    def setUp(self):
        self.authentifier = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
        )

    def test_connection(self):
        # With classic ldap uri (no SSL/TLS)
        self.authentifier.connection()

    @patch("rfl.authentication.ldap.ldap")
    def test_connection_ssl_cert(self, mock_ldap):
        mock_ldap_object = mock_ldap.initialize.return_value

        # With ldap URI, check set_option is not called
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_not_called()

        # With ldaps URI and no CA certificate path, check LDAP server certificate is
        # required and validated with default system OpenSSL CA certificates.
        self.authentifier.uri = urllib.parse.urlparse("ldaps://localhost")
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_REQUIRE_CERT, mock_ldap.OPT_X_TLS_DEMAND
        )
        paths = ssl.get_default_verify_paths()
        if paths.openssl_cafile and os.path.isfile(paths.openssl_cafile):
            mock_ldap_object.set_option.assert_any_call(
                mock_ldap.OPT_X_TLS_CACERTFILE, paths.openssl_cafile
            )
        elif paths.openssl_capath and os.path.isdir(paths.openssl_capath):
            mock_ldap_object.set_option.assert_any_call(
                mock_ldap.OPT_X_TLS_CACERTDIR, paths.openssl_capath
            )
        mock_ldap_object.reset_mock()

        # With CA certificate path, check LDAP server certificate is required and
        # validated with provided CA certificate.
        cert = Path("/dev/null")
        self.authentifier.cacert = cert
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_REQUIRE_CERT, mock_ldap.OPT_X_TLS_DEMAND
        )
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_CACERTFILE, str(cert)
        )
        mock_ldap_object.reset_mock()

    @patch("rfl.authentication.ldap.ldap")
    def test_connection_starttls(self, mock_ldap):
        mock_ldap_object = mock_ldap.initialize.return_value

        # With ldaps URI (and starttls set to False by default), check LDAP server
        # certificate is required and start_tls_s is NOT called.
        self.authentifier.uri = urllib.parse.urlparse("ldaps://localhost")
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_REQUIRE_CERT, mock_ldap.OPT_X_TLS_DEMAND
        )
        mock_ldap_object.start_tls_s.assert_not_called()
        mock_ldap_object.reset_mock()

        # With ldap URI and starttls set to True, check LDAP server certificate is
        # required and start_tls_s is called.
        self.authentifier.uri = urllib.parse.urlparse("ldap://localhost")
        self.authentifier.starttls = True
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_REQUIRE_CERT, mock_ldap.OPT_X_TLS_DEMAND
        )
        mock_ldap_object.start_tls_s.assert_called_once()

    @patch.object(ldap.ldapobject.LDAPObject, "start_tls_s")
    def test_connection_tls_errors(self, mock_start_tls_s):
        self.authentifier.uri = urllib.parse.urlparse("ldap://localhost")
        self.authentifier.starttls = True
        mock_start_tls_s.side_effect = ldap.CONNECT_ERROR("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to connect to LDAP server with STARTTLS: fail$",
        ):
            self.authentifier.connection()
        mock_start_tls_s.side_effect = ldap.SERVER_DOWN("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^LDAP server {self.authentifier.uri.geturl()} is unreachable$",
        ):
            self.authentifier.connection()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_disabled(self, mock_ldap_initialize):
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value

        self.assertEqual(
            self.authentifier._lookup_user_dn("john"),
            f"{self.authentifier.user_name_attribute}=john,"
            f"{self.authentifier.user_base}",
        )
        mock_ldap_object.simple_bind_s.assert_not_called()
        mock_ldap_object.search_s.assert_not_called()
        mock_ldap_object.unbind_s.assert_not_called()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled(self, mock_ldap_initialize):
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.return_value = [
            (
                f"uid=john,ou=admins,{self.authentifier.user_base}",
                {"cn": [b"John Doe"]},
            )
        ]

        self.assertEqual(
            self.authentifier._lookup_user_dn("john"),
            f"uid=john,ou=admins,{self.authentifier.user_base}",
        )
        mock_ldap_object.simple_bind_s.assert_not_called()
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_bind_dn(self, mock_ldap_initialize):
        # define bind dn/password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        self.authentifier.bind_password = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.return_value = [
            (
                f"uid=john,ou=admins,{self.authentifier.user_base}",
                {"cn": [b"John Doe"]},
            )
        ]

        self.assertEqual(
            self.authentifier._lookup_user_dn("john"),
            f"uid=john,ou=admins,{self.authentifier.user_base}",
        )
        mock_ldap_object.simple_bind_s.assert_called_once_with(
            self.authentifier.bind_dn, self.authentifier.bind_password
        )
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_escape_special_chars(self, mock_ldap_initialize):
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        user_with_special_chars = "John Doe(user)"
        escaped_user = ldap.filter.escape_filter_chars(user_with_special_chars)
        mock_ldap_object.search_s.return_value = [
            (
                f"uid={user_with_special_chars},ou=admins,{self.authentifier.user_base}",
                {"cn": [b"John Doe(user)"]},
            )
        ]

        self.assertEqual(
            self.authentifier._lookup_user_dn(user_with_special_chars),
            f"uid={user_with_special_chars},ou=admins,{self.authentifier.user_base}",
        )
        # Verify the filter contains escaped user value
        call_args = mock_ldap_object.search_s.call_args
        search_filter = call_args[0][2]
        # Verify the filter contains escaped user value
        self.assertIn(
            f"({self.authentifier.user_name_attribute}={escaped_user})", search_filter
        )
        # The unescaped value should NOT be in the filter
        self.assertNotIn(
            f"({self.authentifier.user_name_attribute}={user_with_special_chars})",
            search_filter,
        )
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_bind_dn_missing_password(
        self, mock_ldap_initialize
    ):
        # define bind_dn without password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value

        # Check exception is raised when bind_dn is set without bind_password
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Password to authenticate with bind DN uid=read,ou=apps,dc=corp,dc=org "
            r"is required$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_not_called()
        mock_ldap_object.unbind_s.assert_not_called()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_bind_dn_invalid_credentials(
        self, mock_ldap_initialize
    ):
        # define bind dn/password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        self.authentifier.bind_password = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.simple_bind_s.side_effect = ldap.INVALID_CREDENTIALS("fail")

        # Check exception is raised with LDAP fails due to invalid credential
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Invalid bind DN or password$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_not_called()
        mock_ldap_object.unbind_s.assert_not_called()

    @patch.object(ldap, "initialize")
    def test_lookup_user_ldap_server_down_error(self, mock_ldap_initialize):
        # define bind dn/password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        self.authentifier.bind_password = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.simple_bind_s.side_effect = ldap.SERVER_DOWN("fail")

        # Check exception is raised due to LDAP server down
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^LDAP server {self.authentifier.uri.geturl()} is unreachable$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_not_called()
        mock_ldap_object.unbind_s.assert_not_called()

    @patch.object(ldap, "initialize")
    def test_lookup_user_ldap_operations_error(self, mock_ldap_initialize):
        # define bind dn/password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        self.authentifier.bind_password = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.side_effect = ldap.OPERATIONS_ERROR("fail")

        # Check exception is raised due to LDAP server down
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Operations error on user DN lookup: fail$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_not_found(self, mock_ldap_initialize):
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.return_value = []

        # Check exception is raised due to no result found
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to find user john in base ou=people,dc=corp,dc=org$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_too_many_results(self, mock_ldap_initialize):
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.return_value = [
            (
                f"uid=john,ou=admins,{self.authentifier.user_base}",
                {"cn": [b"John Doe"]},
            ),
            (
                f"uid=alice,ou=admins,{self.authentifier.user_base}",
                {"cn": [b"Alice Doe"]},
            ),
        ]
        # Check exception is raised due to too many results found
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Too many users found \(2\) with username john in base "
            r"ou=people,dc=corp,dc=org$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch("rfl.authentication.ldap.ldap")
    def test_login_ok(
        self, mock_ldap, mock_get_groups, mock_get_user_info, mock_lookup_user_dn
    ):
        # setup mocks return values
        mock_get_groups.return_value = ["group1", "group2"]
        mock_get_user_info.return_value = ("John Doe", 42)
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"
        mock_ldap_object = mock_ldap.initialize.return_value

        # call method
        user = self.authentifier.login("john", "SECR3T")

        # verify mock calls
        mock_ldap_object.unbind_s.assert_called_once()
        mock_ldap_object.simple_bind_s.assert_called_once_with(
            "uid=john,ou=people,dc=corp,dc=org", "SECR3T"
        )
        mock_get_user_info.assert_called_once_with(
            mock_ldap_object, "uid=john,ou=people,dc=corp,dc=org"
        )
        mock_get_groups.assert_called_once_with(
            mock_ldap_object, "john", "uid=john,ou=people,dc=corp,dc=org", 42
        )

        # verify return value
        self.assertEqual(user.login, "john")
        self.assertEqual(user.fullname, "John Doe")
        self.assertEqual(user.groups, ["group1", "group2"])

    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch("rfl.authentication.ldap.ldap")
    def test_login_not_in_restricted_group(
        self, mock_ldap, mock_get_groups, mock_get_user_info, mock_lookup_user_dn
    ):
        # setup mocks return values
        self.authentifier.restricted_groups = ["group3", "group4"]
        mock_get_groups.return_value = ["group1", "group2"]
        mock_get_user_info.return_value = ("John Doe", 42)
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"

        # call method
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^User john is not member of restricted groups$",
        ):
            self.authentifier.login("john", "SECR3T")

    def test_login_missing_user_or_password(self):
        with self.assertRaisesRegex(
            LDAPAuthenticationError, "Invalid authentication request"
        ):
            self.authentifier.login("john", None)
            self.authentifier.login(None, "SECR3T")

    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(ldap.ldapobject.LDAPObject, "simple_bind_s")
    def test_login_errors(self, mock_simple_bind_s, mock_lookup_user_dn):
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"
        mock_simple_bind_s.side_effect = ldap.SERVER_DOWN("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^LDAP server {self.authentifier.uri.geturl()} is unreachable$",
        ):
            self.authentifier.login("john", "SECR3T")
        mock_simple_bind_s.side_effect = ldap.INVALID_CREDENTIALS("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError, "^Invalid user or password$"
        ):
            self.authentifier.login("john", "SECR3return_valueT")
        mock_simple_bind_s.side_effect = ldap.UNWILLING_TO_PERFORM("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError, "^LDAP server is unwilling to perform: fail$"
        ):
            self.authentifier.login("john", "SECR3T")

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(LDAPAuthentifier, "_bind")
    @patch.object(LDAPAuthentifier, "connection")
    def test_login_no_bind_lookup_as_user_true(
        self,
        mock_connection,
        mock_bind,
        mock_lookup_user_dn,
        mock_get_user_info,
        mock_get_groups,
    ):
        # setup mocks return values
        mock_get_groups.return_value = ["group1", "group2"]
        mock_get_user_info.return_value = ("John Doe", 42)
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"
        mock_connection.return_value.simple_bind_s.return_value = None

        # if lookup_as_user is False, login() must not call _bind().
        self.authentifier.lookup_as_user = True
        self.authentifier.login("john", "SECR3T")
        mock_bind.assert_not_called()
        mock_connection.return_value.simple_bind_s.assert_called_once_with(
            "uid=john,ou=people,dc=corp,dc=org", "SECR3T"
        )

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(LDAPAuthentifier, "_bind")
    @patch.object(LDAPAuthentifier, "connection")
    def test_login_single_bind_lookup_as_user_false(
        self,
        mock_connection,
        mock_bind,
        mock_lookup_user_dn,
        mock_get_user_info,
        mock_get_groups,
    ):
        # setup mocks return values
        mock_get_groups.return_value = ["group1", "group2"]
        mock_get_user_info.return_value = ("John Doe", 42)
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"

        # if lookup_as_user is False, login() should call _bind() once.
        self.authentifier.lookup_as_user = False
        self.authentifier.login("john", "SECR3T")
        mock_bind.assert_called_once()
        mock_connection.return_value.simple_bind_s.assert_called_once_with(
            "uid=john,ou=people,dc=corp,dc=org", "SECR3T"
        )

    def test_user_info(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            (
                "uid=john,ou=people,dc=corp,dc=org",
                {"cn": [b"John Doe"], "gidNumber": [b"42"]},
            )
        ]
        fullname, gid = self.authentifier._get_user_info(
            connection, "uid=john,ou=people,dc=corp,dc=org"
        )
        self.assertEqual(fullname, "John Doe")
        self.assertEqual(gid, 42)

    def test_user_info_fullname_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # If the user entries in LDAP directory do not contain attributes whose name
        # matches user_fullname_attribute or user_primary_group_attribute, search_s
        # returns a dict with missing keys in the second element of the result tuple.
        #
        # Test missing user_fullname_attribute
        connection.search_s.return_value = [
            (
                "uid=john,ou=people,dc=corp,dc=org",
                {"gidNumber": [b"42"]},
            )
        ]
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to extract user full name with "
            rf"{self.authentifier.user_fullname_attribute} attribute from user "
            r"entries$",
        ):
            self.authentifier._get_user_info(
                connection, "uid=john,ou=people,dc=corp,dc=org"
            )

    def test_user_info_primary_group_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # Test missing gidNumber
        connection.search_s.return_value = [
            (
                "uid=john,ou=people,dc=corp,dc=org",
                {"cn": [b"John Doe"]},
            )
        ]
        with self.assertLogs("rfl", level="WARNING") as log:
            fullname, gid = self.authentifier._get_user_info(
                connection, "uid=john,ou=people,dc=corp,dc=org"
            )
            self.assertEqual(
                [
                    "WARNING:rfl.authentication.ldap:Unable to extract user primary "
                    "group with gidNumber attribute from user entry"
                ],
                log.output,
            )
        self.assertEqual(fullname, "John Doe")
        self.assertIsNone(gid)

    def test_user_info_class_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # If entries with class user_class is not found in user_base subtree,
        # search_s returns an empty list.
        connection.search_s.return_value = []
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^User not found in LDAP with class {self.authentifier.user_class}$",
        ):
            self.authentifier._get_user_info(
                connection, "uid=john,ou=people,dc=corp,dc=org"
            )

    def test_user_info_dn_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # When user DN is not found in LDAP, ldap module raises NO_SUCH_OBJECT
        # exception.
        connection.search_s.side_effect = ldap.NO_SUCH_OBJECT("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^Unable to find user DN uid=john,{self.authentifier.user_base}$",
        ):
            self.authentifier._get_user_info(
                connection, "uid=john,ou=people,dc=corp,dc=org"
            )

    def test_custom_primary_attribute(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        self.authentifier.user_primary_group_attribute = "primaryGroupId"
        dn = "uid=jane,ou=people,dc=corp,dc=org"
        connection.search_s.return_value = [
            (
                dn,
                {
                    "cn": [b"Jane Smith"],
                    self.authentifier.user_primary_group_attribute: [b"42"],
                },
            )
        ]
        fullname, gid = self.authentifier._get_user_info(connection, dn)
        connection.search_s.assert_called_once_with(
            dn,
            ldap.SCOPE_BASE,
            f"(objectClass={self.authentifier.user_class})",
            [
                self.authentifier.user_fullname_attribute,
                self.authentifier.user_primary_group_attribute,
            ],
        )
        self.assertEqual(fullname, "Jane Smith")
        self.assertEqual(gid, 42)

    def test_get_groups(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {"cn": [b"scientists"]}),
            ("cn=biology,ou=groups,dc=corp,dc=org", {"cn": [b"biology"]}),
        ]
        groups = self.authentifier._get_groups(
            connection, "john", "uid=john,ou=people,dc=corp,dc=org", 42
        )
        self.assertEqual(groups, ["scientists", "biology"])

    def test_get_groups_escape_special_chars(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {"cn": [b"scientists"]}),
        ]
        user_name_with_special = "John Doe(user)"
        user_dn_with_special = "uid=John Doe(user),ou=people,dc=corp,dc=org"
        escaped_user_name = ldap.filter.escape_filter_chars(user_name_with_special)
        escaped_user_dn = ldap.filter.escape_filter_chars(user_dn_with_special)
        gid = 42
        groups = self.authentifier._get_groups(
            connection, user_name_with_special, user_dn_with_special, gid
        )
        self.assertEqual(groups, ["scientists"])

        connection.search_s.assert_called_once_with(
            self.authentifier.group_base,
            ldap.SCOPE_SUBTREE,
            f"(&(|(objectClass=posixGroup)(objectClass=groupOfNames))"
            f"(|(memberUid={escaped_user_name})"
            f"(member={escaped_user_dn})(gidNumber={gid})))",
            [self.authentifier.group_name_attribute],
        )

    def test_get_groups_without_gid(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {"cn": [b"scientists"]}),
            ("cn=biology,ou=groups,dc=corp,dc=org", {"cn": [b"biology"]}),
        ]
        user = "john"
        dn = "uid=john,ou=people,dc=corp,dc=org"
        gid = 42
        # First call with gid and check LDAP search filter.
        groups = self.authentifier._get_groups(connection, user, dn, gid)
        self.assertEqual(groups, ["scientists", "biology"])
        connection.search_s.assert_called_once_with(
            self.authentifier.group_base,
            ldap.SCOPE_SUBTREE,
            "(&(|(objectClass=posixGroup)(objectClass=groupOfNames))"
            f"(|(memberUid={user})(member={dn})(gidNumber={gid})))",
            [self.authentifier.group_name_attribute],
        )
        connection.search_s.reset_mock()
        # Then a second call with undefined gid must remove gidNumber from LDAP search
        # filter.
        groups = self.authentifier._get_groups(connection, user, dn, None)
        self.assertEqual(groups, ["scientists", "biology"])
        connection.search_s.assert_called_once_with(
            self.authentifier.group_base,
            ldap.SCOPE_SUBTREE,
            "(&(|(objectClass=posixGroup)(objectClass=groupOfNames))"
            f"(|(memberUid={user})(member={dn})))",
            [self.authentifier.group_name_attribute],
        )

    def test_groups_base_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # When group base DN is not found in LDAP, ldap module raises NO_SUCH_OBJECT
        # exception.
        connection.search_s.side_effect = ldap.NO_SUCH_OBJECT("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^Unable to find group base {self.authentifier.group_base}$",
        ):
            self.authentifier._get_groups(
                connection, "john", "uid=john,ou=people,dc=corp,dc=org", 42
            )

    def test_get_groups_name_attribute_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        self.authentifier.group_name_attribute = "fail"
        # If the group entries in LDAP directory does not contain attributes whose name
        # matches group_name_attribute, search_s returns an empty dict in the second
        # element of the result tuple.
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {}),
            ("cn=biology,ou=groups,dc=corp,dc=org", {}),
        ]
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to extract group name with fail attribute from group entries$",
        ):
            self.authentifier._get_groups(
                connection, "john", "uid=john,ou=people,dc=corp,dc=org", 42
            )

    def test_get_groups_class_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # If entries with one of group_object_classes is not found in group_base
        # subtree, search_s returns an empty list.
        connection.search_s.return_value = []
        with self.assertLogs("rfl.authentication.ldap", level="WARNING") as cm:
            groups = self.authentifier._get_groups(
                connection, "john", "uid=john,ou=people,dc=corp,dc=org", 42
            )
        self.assertEqual(groups, [])
        self.assertEqual(
            cm.output,
            [
                "WARNING:rfl.authentication.ldap:Unable to find groups in LDAP for user"
                " john or gidNumber 42"
            ],
        )
        # Test log message without gid
        with self.assertLogs("rfl.authentication.ldap", level="WARNING") as cm:
            groups = self.authentifier._get_groups(
                connection, "john", "uid=john,ou=people,dc=corp,dc=org", None
            )
        self.assertEqual(groups, [])
        self.assertEqual(
            cm.output,
            [
                "WARNING:rfl.authentication.ldap:Unable to find groups in LDAP for user"
                " john"
            ],
        )

    def test_custom_group_object_classes(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {"cn": [b"scientists"]}),
            ("cn=biology,ou=groups,dc=corp,dc=org", {"cn": [b"biology"]}),
        ]
        group_object_class = "group"
        login = "john"
        gid = 42
        self.authentifier.group_object_classes = [group_object_class]
        groups = self.authentifier._get_groups(
            connection, login, f"uid={login},ou=people,dc=corp,dc=org", gid
        )
        connection.search_s.assert_called_once_with(
            self.authentifier.group_base,
            ldap.SCOPE_SUBTREE,
            f"(&(|(objectClass={group_object_class}))(|(memberUid={login})"
            f"(member=uid={login},ou=people,dc=corp,dc=org)(gidNumber={gid})))",
            [self.authentifier.group_name_attribute],
        )
        self.assertEqual(groups, ["scientists", "biology"])

    def test_in_restricted_groups(self):
        # By default, restricted groups are unset, _in_restricted_groups must return
        # True in all cases.
        self.assertTrue(self.authentifier._in_restricted_groups([]))
        self.assertTrue(self.authentifier._in_restricted_groups(["users", "admins"]))

        # If restricted groups are set, _in_restricted_groups must return if at least
        # one group in argument matches one restricted group.

        # Test with one restricted group
        self.authentifier.restricted_groups = ["admins"]
        self.assertFalse(self.authentifier._in_restricted_groups([]))
        self.assertFalse(
            self.authentifier._in_restricted_groups(["users", "scientists"])
        )
        self.assertTrue(self.authentifier._in_restricted_groups(["admins"]))
        self.assertTrue(
            self.authentifier._in_restricted_groups(["users", "scientists", "admins"])
        )

        # Test with multiple restricted groups
        self.authentifier.restricted_groups = ["admins", "scientists"]
        self.assertFalse(self.authentifier._in_restricted_groups([]))
        self.assertFalse(self.authentifier._in_restricted_groups(["users"]))
        self.assertTrue(
            self.authentifier._in_restricted_groups(["users", "scientists"])
        )
        self.assertTrue(self.authentifier._in_restricted_groups(["admins"]))
        self.assertTrue(
            self.authentifier._in_restricted_groups(["users", "scientists", "admins"])
        )

    def test_list_user_dn(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("uid=john,ou=people,dc=corp,dc=org", {"uid": [b"john"]}),
            ("uid=marie,ou=people,dc=corp,dc=org", {"uid": [b"marie"]}),
        ]
        results = self.authentifier._list_user_dn(connection)
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0], ("john", "uid=john,ou=people,dc=corp,dc=org"))
        self.assertEqual(results[1], ("marie", "uid=marie,ou=people,dc=corp,dc=org"))

    def test_list_user_dn_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # When user DN is not found in LDAP, ldap module raises NO_SUCH_OBJECT
        # exception.
        connection.search_s.side_effect = ldap.NO_SUCH_OBJECT("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to find user base ou=people,dc=corp,dc=org$",
        ):
            self.authentifier._list_user_dn(connection)

    def test_list_user_dn_no_result(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = []
        with self.assertLogs("rfl", level="WARNING") as lc:
            results = self.authentifier._list_user_dn(connection)
            self.assertEqual(
                [
                    "WARNING:rfl.authentication.ldap:Unable to find users in LDAP in "
                    "base ou=people,dc=corp,dc=org subtree"
                ],
                lc.output,
            )
        self.assertEqual(len(results), 0)

    def test_list_user_dn_no_user_name_attribute(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # All results miss the user name attribute, _list_user_dn() is expected to raise
        # LDAPAuthenticationError.
        connection.search_s.return_value = [
            ("uid=john,ou=people,dc=corp,dc=org", {}),
            ("uid=jane,ou=people,dc=corp,dc=org", {}),
        ]
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to extract user uid from user entries$",
        ):
            self.authentifier._list_user_dn(connection)

    def test_list_user_dn_missing_user_name_attribute(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # At least one result has the user name attribute, _list_user_dn() must return
        # these results and log warning message for other dn that miss this attribute.
        connection.search_s.return_value = [
            ("uid=john,ou=people,dc=corp,dc=org", {}),
            ("uid=jane,ou=people,dc=corp,dc=org", {"uid": [b"jane"]}),
        ]
        with self.assertLogs("rfl", level="INFO") as lc:
            results = self.authentifier._list_user_dn(connection)

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0], ("jane", "uid=jane,ou=people,dc=corp,dc=org"))

        # Check warning message to indicate user name attribute has not been found in
        # john user entry has been sent
        self.assertEqual(
            [
                "WARNING:rfl.authentication.ldap:Unable to find uid from user entry "
                "uid=john,ou=people,dc=corp,dc=org"
            ],
            lc.output,
        )

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_list_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch("rfl.authentication.ldap.ldap")
    def test_users(
        self, mock_ldap, mock_get_user_info, mock_list_user_dn, mock_get_groups
    ):
        # Setup mocks return values
        mock_list_user_dn.return_value = [
            ("john", "uid=john,ou=people,dc=corp,dc=org"),
            ("marie", "uid=marie,ou=people,dc=corp,dc=org"),
        ]
        mock_get_user_info.side_effect = [("John Doe", 42), ("Marie Magic", 43)]
        mock_ldap_object = mock_ldap.initialize.return_value

        # Call method (without groups)
        users = self.authentifier.users()

        # Verify mock calls
        mock_ldap_object.unbind_s.assert_called_once()
        mock_list_user_dn.assert_called_once_with(mock_ldap_object)
        mock_get_user_info.assert_called()

        # Verify return value
        self.assertEqual(len(users), 2)
        self.assertEqual(users[0].login, "john")
        self.assertEqual(users[0].fullname, "John Doe")
        self.assertEqual(users[0].groups, [])
        self.assertEqual(users[1].login, "marie")
        self.assertEqual(users[1].fullname, "Marie Magic")
        self.assertEqual(users[1].groups, [])

        # Reset mock and re-inject side effects
        mock_ldap_object.reset_mock()
        mock_list_user_dn.reset_mock()
        mock_get_user_info.reset_mock()
        mock_get_user_info.side_effect = [("John Magic", 45), ("Marie Doe", 46)]
        mock_get_groups.side_effect = [["admin", "users"], ["biology", "users"]]

        # Call method (with groups)
        users = self.authentifier.users(True)

        # Verify mock calls
        mock_ldap_object.unbind_s.assert_called_once()
        mock_list_user_dn.assert_called_once_with(mock_ldap_object)
        mock_get_user_info.assert_called()

        # Verify return value
        self.assertEqual(len(users), 2)
        self.assertEqual(users[0].login, "john")
        self.assertEqual(users[0].fullname, "John Magic")
        self.assertEqual(users[0].groups, ["admin", "users"])
        self.assertEqual(users[1].login, "marie")
        self.assertEqual(users[1].fullname, "Marie Doe")
        self.assertEqual(users[1].groups, ["biology", "users"])

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_list_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch("rfl.authentication.ldap.ldap")
    def test_users_restricted_groups(
        self, mock_ldap, mock_get_user_info, mock_list_user_dn, mock_get_groups
    ):
        self.authentifier.restricted_groups = ["biology"]

        # Setup mocks return values
        mock_list_user_dn.return_value = [
            ("john", "uid=john,ou=people,dc=corp,dc=org"),
            ("marie", "uid=marie,ou=people,dc=corp,dc=org"),
        ]
        mock_get_user_info.side_effect = [("John Magic", 45), ("Marie Doe", 46)]
        mock_get_groups.side_effect = [["admin", "users"], ["biology", "users"]]

        # Call users method with groups retrieval
        with self.assertLogs("rfl", level="DEBUG") as lc:
            users = self.authentifier.users(True)

        # Verify return value. Only user Marie in restricted biology group must be
        # present.
        self.assertEqual(len(users), 1)
        self.assertEqual(users[0].login, "marie")
        self.assertEqual(users[0].fullname, "Marie Doe")
        self.assertEqual(users[0].groups, ["biology", "users"])

        # Check debug message to indicate john is discarded has been sent
        self.assertEqual(
            [
                "DEBUG:rfl.authentication.ldap:Discarding user john not member of "
                "restricted groups"
            ],
            lc.output,
        )

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_list_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch("rfl.authentication.ldap.ldap")
    def test_users_bind_dn(
        self, mock_ldap, mock_get_user_info, mock_list_user_dn, mock_get_groups
    ):
        self.authentifier.bind_dn = "uid=hey,ou=people,dc=corp,dc=org"
        self.authentifier.bind_password = "secr3t"

        mock_ldap_object = mock_ldap.initialize.return_value

        # Call method
        self.authentifier.users()

        # Check simple_bind_s is called when bind_dn and bind_password are defined
        mock_ldap_object.simple_bind_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch("rfl.authentication.ldap.ldap")
    def test_users_bind_dn_missing_password(self, mock_ldap):
        self.authentifier.bind_dn = "uid=hey,ou=people,dc=corp,dc=org"

        # Check exception is raised when bind_dn is set without bind_password
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Password to authenticate with bind DN uid=hey,ou=people,dc=corp,dc=org "
            r"is required$",
        ):
            self.authentifier.users()

    @patch.object(ldap, "initialize")
    def test_users_bind_dn_invalid_credentials(self, mock_ldap_initialize):
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.simple_bind_s.side_effect = ldap.INVALID_CREDENTIALS("fail")

        self.authentifier.bind_dn = "uid=hey,ou=people,dc=corp,dc=org"
        self.authentifier.bind_password = "secr3t"

        # Check exception is raised with LDAP fails due to invalid credential
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Invalid bind DN or password$",
        ):
            self.authentifier.users()

    @patch.object(ldap.ldapobject.LDAPObject, "search_s")
    def test_users_ldap_server_down_error(self, mock_search_s):
        mock_search_s.side_effect = ldap.SERVER_DOWN("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^LDAP server {self.authentifier.uri.geturl()} is unreachable$",
        ):
            self.authentifier.users()

    @patch.object(ldap.ldapobject.LDAPObject, "search_s")
    def test_users_ldap_operations_error(self, mock_search_s):
        mock_search_s.side_effect = ldap.OPERATIONS_ERROR("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Operations error on users search: fail$",
        ):
            self.authentifier.users()


class TestLDAPAuthentifierInit(unittest.TestCase):
    def test_lookup_as_user_auto_bind_dn(self):
        # lookup_as_user is None, bind_dn and bind_password are set, should do lookup
        # with service credentials.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            bind_dn="uid=read,ou=apps,dc=corp,dc=org",
            bind_password="SECR3T",
            lookup_as_user=None,
        )
        self.assertFalse(auth.lookup_as_user)

    def test_lookup_as_user_auto_no_bind_user(self):
        # lookup_as_user is None, bind_dn and bind_password are not set, should do
        # lookup as user.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            bind_dn=None,
            bind_password=None,
            lookup_as_user=None,
        )
        self.assertTrue(auth.lookup_as_user)

    def test_lookup_as_user_auto_no_bind_dn(self):
        # lookup_as_user is None, bind_dn is not set, should do lookup as user.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            bind_dn=None,
            bind_password="SECR3T",
            lookup_as_user=None,
        )
        self.assertTrue(auth.lookup_as_user)

    def test_lookup_as_user_auto_no_bind_password(self):
        # lookup_as_user is None, bind_password is not set, should do lookup as user.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            bind_dn="uid=read,ou=apps,dc=corp,dc=org",
            bind_password=None,
            lookup_as_user=None,
        )
        self.assertTrue(auth.lookup_as_user)

    def test_lookup_as_user_enabled(self):
        # lookup_as_user is True, should do lookup as user.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            lookup_as_user=True,
        )
        self.assertTrue(auth.lookup_as_user)

    def test_lookup_as_user_disabled(self):
        # lookup_as_user is False, should do lookup with service credentials.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            lookup_as_user=False,
        )
        self.assertFalse(auth.lookup_as_user)



# Copyright (c) 2025 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
import logging

from rfl.log import setup_logger, enforce_debug
from rfl.log.formatters import DaemonFormatter


class TestLogFilterer(unittest.TestCase):
    def assertLogRecordNotFiltered(self, handler, attrdict):
        record = logging.makeLogRecord(attrdict)
        try:
            self.assertTrue(handler.filter(record))
        except AssertionError:
            self.assertIsInstance(handler.filter(record), logging.LogRecord)

    def assertLogRecordFiltered(self, handler, attrdict):
        self.assertFalse(handler.filter(logging.makeLogRecord(attrdict)), False)


class TestSetupLogger(TestLogFilterer):
    def test_setup(self):
        setup_logger()
        # get root logger
        logger = logging.getLogger()
        # check root logger level
        self.assertEqual(logger.level, logging.INFO)
        # get stream handler and check its level
        handler = logger.handlers[-1]
        self.assertEqual(handler.level, logging.INFO)
        # check formatter
        self.assertIsInstance(handler.formatter, DaemonFormatter)
        # check custom filter
        self.assertEqual(len(handler.filters), 1)

    def test_setup_debug(self):
        setup_logger(debug=True)
        # get root logger
        logger = logging.getLogger()
        # check root logger level
        self.assertEqual(logger.level, logging.DEBUG)
        # get stream handler and check its level
        handler = logger.handlers[-1]
        self.assertEqual(handler.level, logging.DEBUG)

    def test_setup_warning(self):
        setup_logger(level=logging.WARNING)
        logger = logging.getLogger()
        self.assertEqual(logger.level, logging.WARNING)
        handler = logger.handlers[-1]
        self.assertEqual(handler.level, logging.WARNING)

    def test_setup_error(self):
        setup_logger(level=logging.ERROR)
        logger = logging.getLogger()
        self.assertEqual(logger.level, logging.ERROR)
        handler = logger.handlers[-1]
        self.assertEqual(handler.level, logging.ERROR)

    def test_setup_level_string(self):
        setup_logger(level="WARNING")
        logger = logging.getLogger()
        self.assertEqual(logger.level, logging.WARNING)
        handler = logger.handlers[-1]
        self.assertEqual(handler.level, logging.WARNING)

    def test_setup_level_overrides_debug(self):
        setup_logger(level=logging.WARNING, debug=True)
        logger = logging.getLogger()
        self.assertEqual(logger.level, logging.WARNING)
        handler = logger.handlers[-1]
        self.assertEqual(handler.level, logging.WARNING)

    def test_setup_invalid_level(self):
        with self.assertRaisesRegex(ValueError, r"Unknown log level: NOTALEVEL"):
            setup_logger(level="NOTALEVEL")

    def test_setup_invalid_level_type(self):
        with self.assertRaisesRegex(TypeError, r"Unsupported log level type: float"):
            setup_logger(level=1.5)

    def test_setup_filter(self):
        setup_logger(log_flags=["rfl"])
        # get root logger
        logger = logging.getLogger()
        # get stream handler
        handler = logger.handlers[-1]
        self.assertLogRecordNotFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.INFO, "msg": "test"}
        )
        self.assertLogRecordFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.DEBUG, "msg": "test"}
        )
        self.assertLogRecordFiltered(
            handler, {"name": "fail.pkg", "levelno": logging.WARNING, "msg": "test"}
        )

    def test_setup_filter_debug_flags(self):
        setup_logger(debug=True, log_flags=["rfl"], debug_flags=["test"])
        # get root logger
        logger = logging.getLogger()
        # get stream handler
        handler = logger.handlers[-1]
        self.assertLogRecordNotFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.INFO, "msg": "test"}
        )
        self.assertLogRecordNotFiltered(
            handler, {"name": "test.pkg", "levelno": logging.DEBUG, "msg": "test"}
        )
        self.assertLogRecordFiltered(
            handler, {"name": "fail.pkg", "levelno": logging.WARNING, "msg": "test"}
        )
        self.assertLogRecordFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.DEBUG, "msg": "test"}
        )

    def test_setup_filter_debug_all(self):
        setup_logger(debug=True, log_flags=["ALL"], debug_flags=["ALL"])
        # get root logger
        logger = logging.getLogger()
        # get stream handler
        handler = logger.handlers[-1]
        self.assertLogRecordNotFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.INFO, "msg": "test"}
        )
        self.assertLogRecordNotFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.DEBUG, "msg": "test"}
        )
        self.assertLogRecordNotFiltered(
            handler, {"name": "fail.pkg", "levelno": logging.WARNING, "msg": "test"}
        )
        self.assertLogRecordNotFiltered(
            handler, {"name": "fail.pkg", "levelno": logging.DEBUG, "msg": "test"}
        )

    def test_setup_clears_handlers(self):
        setup_logger()
        setup_logger()
        logger = logging.getLogger()
        self.assertEqual(len(logger.handlers), 1)

    def test_setup_clear_false_stacks_handlers(self):
        root_logger = logging.getLogger()
        for handler in root_logger.handlers[:]:
            root_logger.removeHandler(handler)
            handler.close()
        setup_logger(clear=False)
        setup_logger(clear=False)
        self.assertEqual(len(root_logger.handlers), 2)


class TestEnforceLogger(TestLogFilterer):
    def setUp(self):
        setup_logger()

    def test_enforce_debug_flags(self):
        enforce_debug(log_flags=["rfl"], debug_flags=["test"])
        # get root logger
        logger = logging.getLogger()
        # get stream handler
        handler = logger.handlers[-1]
        self.assertLogRecordNotFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.INFO, "msg": "test"}
        )
        self.assertLogRecordNotFiltered(
            handler, {"name": "test.pkg", "levelno": logging.DEBUG, "msg": "test"}
        )
        self.assertLogRecordFiltered(
            handler, {"name": "fail.pkg", "levelno": logging.WARNING, "msg": "test"}
        )
        self.assertLogRecordFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.DEBUG, "msg": "test"}
        )

    def test_enforce_debug_all(self):
        enforce_debug(log_flags=["ALL"], debug_flags=["ALL"])
        # get root logger
        logger = logging.getLogger()
        # get stream handler
        handler = logger.handlers[-1]
        self.assertLogRecordNotFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.INFO, "msg": "test"}
        )
        self.assertLogRecordNotFiltered(
            handler, {"name": "rfl.pkg", "levelno": logging.DEBUG, "msg": "test"}
        )
        self.assertLogRecordNotFiltered(
            handler, {"name": "fail.pkg", "levelno": logging.WARNING, "msg": "test"}
        )
        self.assertLogRecordNotFiltered(
            handler, {"name": "fail.pkg", "levelno": logging.DEBUG, "msg": "test"}
        )



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import typing as t
import secrets
from datetime import datetime, timezone, timedelta
import logging
from pathlib import Path

try:
    import jwt
except ImportError as err:
    raise ImportError("PyJWT is required for RFL JWT Authentication") from err

from .user import AuthenticatedUser
from .errors import (
    JWTPrivateKeyGeneratorError,
    JWTPrivateKeyLoaderError,
    JWTDecodeError,
    JWTEncodeError,
)

logger = logging.getLogger(__name__)


def jwt_gen_key(path: Path):
    """Generate JWT private key with random data. Raise JWTPrivateKeyGeneratorError in
    case of error."""
    logger.info("Generating JWT private key file %s", path)
    try:
        with open(path, "w+") as fh:
            fh.write(secrets.token_hex(32))
    except (FileNotFoundError, NotADirectoryError, PermissionError) as err:
        raise JWTPrivateKeyGeneratorError(
            f"Error while generating JWT key {path}: {err}"
        ) from err
    path.chmod(0o400)  # restrict access to encryption key


def jwt_validate_expiration(token: str):
    """Load JWT and check its expiration but not its signature, and its claimset. This
    can be useful to validate JWT on client side before using it, without access to
    its private signature key."""
    try:
        return jwt.decode(
            token, options={"verify_signature": False, "verify_exp": True}
        )
    except jwt.ExpiredSignatureError as err:
        raise JWTDecodeError("Token is expired") from err
    except jwt.exceptions.DecodeError as err:
        raise JWTDecodeError(f"Unable to decode token: {str(err)}") from err


class JWTPrivateKeyLoader:
    pass


class JWTPrivateKeyFileLoader(JWTPrivateKeyLoader):
    def __init__(
        self,
        path: Path = None,
        value: str = None,
        create: bool = False,
        create_parent: bool = False,
        text: bool = False,
    ):
        self.text = text
        if path is not None:
            logger.debug("Loading JWT private key from path %s", path)
            self.path = path
            self.key = None
            self._load_path(create, create_parent)
        elif value is not None:
            logger.debug("Loading JWT private key from value")
            self.path = None
            self.key = value
        else:
            raise JWTPrivateKeyLoaderError(
                "Either key value or a path must be given to load JWT private key"
            )

    def _load_path(self, create: bool, create_parent: bool) -> None:
        """Load the JWT private key from path. If create argument is True, the private
        key is randomly generated if not present. If create_parent is True, the parent
        directory of the private key is also created if not present. Raise
        JWTPrivateKeyLoaderError if create_parent is False and private parent directory
        does not exist or if create is False and private key file does not exist."""
        # Create tokens directory if missing
        if not self.path.parent.exists():
            if create_parent:
                logger.info("Creating JWT private key directory %s", self.path.parent)
                try:
                    self.path.parent.mkdir()
                    self.path.parent.chmod(0o755)  # be umask agnostic
                except PermissionError as err:
                    raise JWTPrivateKeyLoaderError(
                        "Permission denied to create key parent directory "
                        f"{self.path.parent}"
                    ) from err
            else:
                raise JWTPrivateKeyLoaderError(
                    f"Token private key parent directory {self.path.parent} not found"
                )
        # Generate instance tokens encryption key file if missing
        if not self.path.exists():
            if create:
                jwt_gen_key(self.path)
            else:
                raise JWTPrivateKeyLoaderError(
                    f"Token private key file {self.path} not found"
                )
        # Load the instance tokens encryption key
        try:
            mode = "rb"
            if self.text:
                mode = "r"
            with open(self.path, mode) as fh:
                self.key = fh.read()
        except PermissionError as err:
            raise JWTPrivateKeyLoaderError(
                f"Permission error to access private key file {self.path}"
            ) from err
        except UnicodeDecodeError as err:
            raise JWTPrivateKeyLoaderError(
                f"Unable to decode private key file {self.path}: {err}"
            ) from err
        if not len(self.key):
            raise JWTPrivateKeyLoaderError(f"Key loaded from file {self.path} is empty")


class JWTBaseManager:
    def __init__(self, algorithm: str, loader: JWTPrivateKeyLoader):
        self.algorithm = algorithm
        self.key = loader.key

    def decode(self, token, **claimset) -> AuthenticatedUser:
        """Decode the given token with the encryption key an returns the user of
        this token."""
        try:
            return jwt.decode(token, self.key, algorithms=[self.algorithm], **claimset)
        except jwt.InvalidSignatureError as err:
            raise JWTDecodeError("Token signature is invalid") from err
        except jwt.ExpiredSignatureError as err:
            raise JWTDecodeError("Token is expired") from err
        except jwt.InvalidAudienceError as err:
            raise JWTDecodeError("Token audience is invalid") from err
        except jwt.exceptions.DecodeError as err:
            raise JWTDecodeError(f"Unable to decode token: {str(err)}") from err

    def generate(
        self, duration: int, claimset: t.Optional[t.Dict[str, t.Any]] = {}
    ) -> str:
        try:
            token = jwt.encode(
                {
                    "iat": datetime.now(tz=timezone.utc),
                    "exp": datetime.now(tz=timezone.utc) + timedelta(days=duration),
                    **claimset,
                },
                self.key,
                algorithm=self.algorithm,
            )
        except NotImplementedError as err:
            raise JWTEncodeError(f"JWT token encode error: {str(err)}") from err
        # For compatibility with PyJWT < 2
        if isinstance(token, bytes):
            return token.decode()
        return token


class JWTManager(JWTBaseManager):
    def __init__(self, audience: str, algorithm: str, loader: JWTPrivateKeyLoader):
        super().__init__(algorithm, loader)
        self.audience = audience

    def decode(self, token) -> AuthenticatedUser:
        """Decode the given token with the encryption key an returns the user of
        this token."""
        try:
            payload = super().decode(token, audience=self.audience)
        except jwt.InvalidAudienceError as err:
            raise JWTDecodeError("Token audience is invalid") from err
        return AuthenticatedUser(login=payload["sub"], groups=payload["groups"])

    def generate(self, user: AuthenticatedUser, duration: int) -> str:
        """Returns a JWT token for the given user, signed with the encryption
        key, for the configured audience and valid for the given duration."""
        return super().generate(
            duration,
            {
                "aud": self.audience,
                "sub": user.login,
                "groups": user.groups,
            },
        )

    @classmethod
    def key(
        cls,
        audience: str,
        algorithm: str,
        path: Path,
        create: bool = False,
        create_parent: bool = False,
    ):
        return cls(
            audience,
            algorithm,
            JWTPrivateKeyFileLoader(
                path=path, create=create, create_parent=create_parent
            ),
        )



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import textwrap
import unittest
import warnings

from rfl.authentication.user import AuthenticatedUser, AnonymousUser
from rfl.permissions.rbac import (
    RBACPolicyRole,
    RBACPolicyDefinitionYAMLLoader,
    RBACPolicyRolesIniLoader,
    RBACPolicyManager,
)
from rfl.permissions.errors import (
    RBACPolicyDefinitionLoadError,
    RBACPolicyRolesLoadError,
)

VALID_DEFINITION = """
actions:
  view-users: View all users information
  add-users: Add users to the system
  view-tasks: View all tasks on the system
  launch-tasks: Launch tasks on the system
  edit-tasks:
    description: Edit all submitted tasks
"""

DEFINITION_DEPRECATED = """
actions:
  old-action:
    description: Old action
    deprecated: true
    replaced_by: new-action
  old-no-replace:
    description: Old action without replacement
    deprecated: true
  new-action:
    description: New action
"""

VALID_ROLES = """
[roles]
anonymous
# All authenticated users have the base role
base=ALL
user=@users,mike,lisa
operator=lisa
administrator=john,@admins

# The anonymous role is permitted to perform all view-* actions.
[anonymous]
actions=view-tasks

[base]
actions=view-users

# "user" role is permitted to perform all anonymous actions and launch-tasks action.
[user]
actions=launch-tasks

[operator]
actions=edit-tasks

# "administrator" role is permitted to perform all anonymous and user actions, augmented
# with edit-tasks and add-users actions.
[administrator]
actions=@user,@operator,add-users
"""


class TestRBACPolicyDefinitionYAMLLoader(unittest.TestCase):
    def test_load_valid_definition(self):
        loader = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        self.assertEqual(
            set(loader.actions.keys()),
            {"view-users", "add-users", "view-tasks", "launch-tasks", "edit-tasks"},
        )

    def test_load_invalid_yaml(self):
        with self.assertRaisesRegex(
            RBACPolicyDefinitionLoadError,
            r"^Invalid YAML policy definition: while parsing a block mapping",
        ):
            RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION + "\n fail")

    def test_load_invalid_actions_key(self):
        raw_definition = "fail:" + VALID_DEFINITION[6:]
        with self.assertRaisesRegex(
            RBACPolicyDefinitionLoadError,
            r"^Actions key not found in YAML policy definition$",
        ):
            RBACPolicyDefinitionYAMLLoader(raw=raw_definition)

    def test_load_invalid_actions_content(self):
        raw_definition = "actions: fail"
        with self.assertRaisesRegex(
            RBACPolicyDefinitionLoadError,
            r"^Actions definition must be a mapping of action names to definitions$",
        ):
            RBACPolicyDefinitionYAMLLoader(raw=raw_definition)

    def test_load_missing_description(self):
        definition = textwrap.dedent("""
            actions:
              old-action:
                deprecated: true
            """)
        with self.assertRaisesRegex(
            RBACPolicyDefinitionLoadError,
            r"^Description is mandatory for action old-action$",
        ):
            RBACPolicyDefinitionYAMLLoader(raw=definition)

    def test_load_invalid_deprecated_not_boolean(self):
        definition = textwrap.dedent("""
            actions:
              old-action:
                description: Old action
                deprecated: "true"
            """)
        with self.assertRaisesRegex(
            RBACPolicyDefinitionLoadError,
            r"^Invalid deprecated value for action old-action$",
        ):
            RBACPolicyDefinitionYAMLLoader(raw=definition)

    def test_load_invalid_replaced_by_list(self):
        definition = textwrap.dedent("""
            actions:
              old-action:
                description: Old action
                deprecated: true
                replaced_by:
                  - a
                  - b
            """)
        with self.assertRaisesRegex(
            RBACPolicyDefinitionLoadError,
            r"^Invalid replaced_by value for action old-action$",
        ):
            RBACPolicyDefinitionYAMLLoader(raw=definition)

    def test_load_invalid_replaced_by_unknown(self):
        definition = textwrap.dedent("""
            actions:
              old-action:
                description: Old action
                deprecated: true
                replaced_by: missing
            """)
        with self.assertRaisesRegex(
            RBACPolicyDefinitionLoadError,
            r"^Replacement action missing for old-action not found in policy "
            r"definition$",
        ):
            RBACPolicyDefinitionYAMLLoader(raw=definition)

    def test_load_legacy_string_description(self):
        loader = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        expected = {
            "view-users": "View all users information",
            "add-users": "Add users to the system",
            "view-tasks": "View all tasks on the system",
            "launch-tasks": "Launch tasks on the system",
            "edit-tasks": "Edit all submitted tasks",
        }
        self.assertEqual(
            {action: meta.description for action, meta in loader.actions.items()},
            expected,
        )

    def test_apply_non_deprecated_action(self):
        loader = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        result = loader.apply("view-users")
        self.assertEqual(result, {"view-users"})

    def test_apply_deprecated_action_without_replacement(self):
        loader = RBACPolicyDefinitionYAMLLoader(raw=DEFINITION_DEPRECATED)
        with self.assertWarns(UserWarning) as warning:
            result = loader.apply("old-no-replace")
        self.assertEqual(result, set())
        self.assertIn("Action old-no-replace is deprecated", str(warning.warning))

    def test_apply_deprecated_action_with_replacement(self):
        loader = RBACPolicyDefinitionYAMLLoader(raw=DEFINITION_DEPRECATED)
        with self.assertWarns(UserWarning) as warning:
            result = loader.apply("old-action")
        self.assertEqual(result, {"new-action"})
        self.assertIn("Action old-action is deprecated", str(warning.warning))
        self.assertIn("use new-action instead", str(warning.warning))

    def test_apply_deprecated_action_recursive_replacement(self):
        definition = textwrap.dedent("""
            actions:
              old-action:
                description: Old action
                deprecated: true
                replaced_by: intermediate-action
              intermediate-action:
                description: Intermediate action
                deprecated: true
                replaced_by: final-action
              final-action:
                description: Final action
            """)
        loader = RBACPolicyDefinitionYAMLLoader(raw=definition)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = loader.apply("old-action")
        self.assertEqual(result, {"final-action"})
        # Should warn about both deprecated actions
        self.assertEqual(len(w), 2, f"Expected 2 warnings, got {len(w)}")
        warnings_messages = [str(warning.message) for warning in w]
        self.assertTrue(
            any("old-action" in msg for msg in warnings_messages),
            f"Expected warning about old-action in {warnings_messages}",
        )
        self.assertTrue(
            any("intermediate-action" in msg for msg in warnings_messages),
            f"Expected warning about intermediate-action in {warnings_messages}",
        )


class TestRBACPolicyRolesIniLoader(unittest.TestCase):
    def test_load_valid_roles(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(definition=definition, raw=VALID_ROLES)
        self.assertEqual(len(loader.roles), 5)
        for role in loader.roles:
            self.assertIsInstance(role, RBACPolicyRole)
            self.assertIn(
                role.name, ["anonymous", "base", "user", "operator", "administrator"]
            )
            if role.name == "user":
                self.assertEqual(role.actions, {"launch-tasks"})
                self.assertEqual(role.members, {"mike", "@users", "lisa"})
            elif role.name == "administrator":
                self.assertEqual(
                    role.actions, {"launch-tasks", "edit-tasks", "add-users"}
                )
                self.assertEqual(role.members, {"john", "@admins"})

    def test_load_roles_missing_roles_section(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(
            definition=definition, raw=VALID_ROLES, preload=False
        )
        loader.content.remove_section("roles")
        with self.assertRaisesRegex(
            RBACPolicyRolesLoadError,
            r"^Section roles to define roles members is not defined in INI content$",
        ):
            loader._load()

    def test_load_roles_missing_role_section(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(
            definition=definition, raw=VALID_ROLES, preload=False
        )
        loader.content.remove_section("administrator")
        with self.assertRaisesRegex(
            RBACPolicyRolesLoadError,
            r"^Section to define actions for role administrator is not defined in INI "
            r"content$",
        ):
            loader._load()

    def test_load_roles_missing_role_option(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(
            definition=definition, raw=VALID_ROLES, preload=False
        )
        loader.content.remove_option("administrator", "actions")
        with self.assertRaisesRegex(
            RBACPolicyRolesLoadError,
            r"^Option to define actions for role administrator is not defined in INI "
            r"content$",
        ):
            loader._load()

    def test_load_roles_undefined_action(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(
            definition=definition, raw=VALID_ROLES, preload=False
        )
        loader.content.set("administrator", "actions", "fail")
        with self.assertRaisesRegex(
            RBACPolicyRolesLoadError,
            r"^Action fail not found in policy definition$",
        ):
            loader._load()

    def test_load_roles_undefined_expand(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(
            definition=definition, raw=VALID_ROLES, preload=False
        )
        loader.content.set("administrator", "actions", "@fail")
        with self.assertRaisesRegex(
            RBACPolicyRolesLoadError,
            r"^Unable to inherit actions from role fail not found in policy$",
        ):
            loader._load()

    def test_load_roles_deprecated_replaced(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=DEFINITION_DEPRECATED)
        roles = textwrap.dedent("""
            [roles]
            user=ALL

            [user]
            actions=old-action
            """)
        with self.assertWarns(UserWarning):
            loader = RBACPolicyRolesIniLoader(definition=definition, raw=roles)
        # old-action stripped, new-action added
        user_role = next(role for role in loader.roles if role.name == "user")
        self.assertEqual(user_role.actions, {"new-action"})

    def test_load_roles_deprecated_without_replacement(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=DEFINITION_DEPRECATED)
        roles = textwrap.dedent("""
            [roles]
            user=ALL

            [user]
            actions=old-no-replace
            """)
        with self.assertWarns(UserWarning):
            loader = RBACPolicyRolesIniLoader(definition=definition, raw=roles)
        user_role = next(role for role in loader.roles if role.name == "user")
        self.assertEqual(user_role.actions, set())

    def test_load_roles_inheritance_with_deprecated(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=DEFINITION_DEPRECATED)
        roles = textwrap.dedent("""
            [roles]
            base=ALL
            user=ALL

            [base]
            actions=old-action

            [user]
            actions=@base
            """)
        with self.assertWarns(UserWarning):
            loader = RBACPolicyRolesIniLoader(definition=definition, raw=roles)
        base_role = next(role for role in loader.roles if role.name == "base")
        user_role = next(role for role in loader.roles if role.name == "user")
        self.assertEqual(base_role.actions, {"new-action"})
        self.assertEqual(user_role.actions, {"new-action"})


class TestRBACPolicyManager(unittest.TestCase):
    def test_manager_init(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(definition=definition, raw=VALID_ROLES)
        manager = RBACPolicyManager(loader)
        self.assertTrue(manager.allow_anonymous)

    def test_manager_disable_anonymous(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(definition=definition, raw=VALID_ROLES)
        manager = RBACPolicyManager(loader)
        self.assertTrue(manager.allow_anonymous)
        manager.disable_anonymous()
        self.assertFalse(manager.allow_anonymous)

    def test_manager_load_wo_anonymous(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(
            definition=definition, raw=VALID_ROLES, preload=False
        )
        loader.content.remove_option("roles", "anonymous")
        loader._load()
        manager = RBACPolicyManager(loader)
        self.assertFalse(manager.allow_anonymous)

    def test_manager_roles_actions(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(definition=definition, raw=VALID_ROLES)
        manager = RBACPolicyManager(loader)
        self.assertEqual(
            manager.roles_actions(AuthenticatedUser(login="FAKE", groups=["users"])),
            (
                {"user", "base"},
                {"launch-tasks", "view-users"},
            ),
        )
        self.assertEqual(
            manager.roles_actions(AuthenticatedUser(login="lisa", groups=["FAKE"])),
            (
                {"operator", "base", "user"},
                {"launch-tasks", "view-users", "edit-tasks"},
            ),
        )
        self.assertEqual(
            manager.roles_actions(AuthenticatedUser(login="FAKE", groups=["admins"])),
            (
                {"base", "administrator"},
                {"edit-tasks", "add-users", "launch-tasks", "view-users"},
            ),
        )

    def test_manager_roles_actions_unknown_user(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(
            definition=definition, raw=VALID_ROLES, preload=False
        )
        manager = RBACPolicyManager(loader)
        # Remove the base role for all authenticated users so unknown user could not be
        # associated to any role defined in policy.
        loader.content.remove_option("roles", "base")
        loader._load()
        manager = RBACPolicyManager(loader)
        self.assertEqual(
            manager.roles_actions(AuthenticatedUser(login="FAKE", groups=["unknown"])),
            (set(), set()),
        )

    def test_manager_roles_actions_anonymous(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(definition=definition, raw=VALID_ROLES)
        manager = RBACPolicyManager(loader)
        self.assertEqual(
            manager.roles_actions(AnonymousUser()),
            ({"anonymous"}, {"view-tasks"}),
        )

    def test_manager_roles_actions_anonymous_disabled(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(
            definition=definition, raw=VALID_ROLES, preload=False
        )
        loader.content.remove_option("roles", "anonymous")
        loader._load()
        manager = RBACPolicyManager(loader)
        self.assertFalse(manager.allow_anonymous)
        self.assertEqual(
            manager.roles_actions(AnonymousUser()),
            (set(), set()),
        )

    def test_manager_allowed_anonymous_action(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(definition=definition, raw=VALID_ROLES)
        manager = RBACPolicyManager(loader)
        self.assertEqual(manager.allowed_anonymous_action("view-tasks"), True)
        self.assertEqual(manager.allowed_anonymous_action("view-users"), False)
        self.assertEqual(manager.allowed_anonymous_action("launch-tasks"), False)

    def test_manager_allowed_anonymous_user_action(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(definition=definition, raw=VALID_ROLES)
        manager = RBACPolicyManager(loader)
        user = AnonymousUser()
        self.assertEqual(manager.allowed_user_action(user, "view-tasks"), True)
        self.assertEqual(manager.allowed_user_action(user, "view-users"), False)
        self.assertEqual(manager.allowed_user_action(user, "launch-tasks"), False)

    def test_manager_allowed_user_action(self):
        definition = RBACPolicyDefinitionYAMLLoader(raw=VALID_DEFINITION)
        loader = RBACPolicyRolesIniLoader(definition=definition, raw=VALID_ROLES)
        manager = RBACPolicyManager(loader)
        # Members of users group must have access to base and user roles actions but not
        # to anonymous and operator role actions.
        user = AuthenticatedUser(login="FAKE", groups=["users"])
        self.assertEqual(manager.allowed_user_action(user, "view-users"), True)
        self.assertEqual(manager.allowed_user_action(user, "launch-tasks"), True)
        self.assertEqual(manager.allowed_user_action(user, "view-tasks"), False)
        self.assertEqual(manager.allowed_user_action(user, "edit-tasks"), False)

        # Mike (user) must have access to user role actions but not to operator role
        # actions.
        user = AuthenticatedUser(login="mike", groups=["FAKE"])
        self.assertEqual(manager.allowed_user_action(user, "launch-tasks"), True)
        self.assertEqual(manager.allowed_user_action(user, "edit-tasks"), False)

        # Members of admins group must have access to all actions
        user = AuthenticatedUser(login="FAKE", groups=["admins"])
        self.assertEqual(manager.allowed_user_action(user, "view-users"), True)
        self.assertEqual(manager.allowed_user_action(user, "edit-tasks"), True)
        self.assertEqual(manager.allowed_user_action(user, "add-users"), True)

        # John (admin) must have access to all actions except those not declared.
        user = AuthenticatedUser(login="john", groups=["FAKE"])
        self.assertEqual(manager.allowed_user_action(user, "view-users"), True)
        self.assertEqual(manager.allowed_user_action(user, "fail"), False)

        # Lisa (user + operator) must have access to base, user and operator roles
        # actions but not to anonymous and admin role actions.
        user = AuthenticatedUser(login="lisa", groups=["users"])
        self.assertEqual(manager.allowed_user_action(user, "edit-tasks"), True)
        self.assertEqual(manager.allowed_user_action(user, "view-users"), True)
        self.assertEqual(manager.allowed_user_action(user, "launch-tasks"), True)
        self.assertEqual(manager.allowed_user_action(user, "view-tasks"), False)
        self.assertEqual(manager.allowed_user_action(user, "add-users"), False)



# Copyright (c) 2024 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
import asyncio

from rfl.core.asyncio import asyncio_run


async def async_success():
    pass


async def async_fail():
    raise RuntimeError("fail")


async def async_sleep():
    await asyncio.sleep(0.5)


async def multiple_sleeps():
    async_sleep()
    async_sleep()
    async_sleep()


class TestAsyncIO(unittest.TestCase):
    def test_run_success(self):
        asyncio_run(async_success())

    def test_run_failure(self):
        with self.assertRaisesRegex(RuntimeError, "^fail$"):
            asyncio_run(async_fail())

    def test_run_sleep(self):
        asyncio_run(multiple_sleeps())



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Union, List


class AuthenticatedUser:
    def __init__(
        self, login: str, fullname: Union[str, None] = None, groups: List[str] = []
    ):
        self.login = login
        self.fullname = fullname
        self.groups = groups

    def __str__(self) -> str:
        return f"{self.login} ({self.fullname or '∅'}) [{', '.join(self.groups)}]"

    def is_anonymous(self):
        """Return True if user has the anonymous login."""
        return self.login == AnonymousUser.LOGIN


class AnonymousUser(AuthenticatedUser):
    LOGIN = "anonymous"
    FULLNAME = "anonymous"

    def __init__(self):
        return super().__init__(AnonymousUser.LOGIN, AnonymousUser.FULLNAME)



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import os
from pathlib import Path
import shutil


def main():
    """Install setup.py script provided by RFL.build package in scripts/ subfolder in
    the current working directory."""
    orig = Path(os.path.realpath(__file__)).parent.joinpath("scripts", "setup")
    dest = Path(os.getcwd()).joinpath("setup.py")
    print(f"Copying file {orig} to {dest}")
    shutil.copy(orig, dest)



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from pathlib import Path
import stat
import tempfile
import os
import unittest

from rfl.authentication.user import AuthenticatedUser, AnonymousUser
from rfl.authentication.jwt import (
    jwt_gen_key,
    jwt_validate_expiration,
    JWTPrivateKeyFileLoader,
    JWTBaseManager,
    JWTManager,
)
from rfl.authentication.errors import (
    JWTPrivateKeyGeneratorError,
    JWTPrivateKeyLoaderError,
    JWTEncodeError,
    JWTDecodeError,
)

PRIVATE_KEY = b"TEST_PRIVATE_KEY"


class TestJWTGenKey(unittest.TestCase):
    def test_gen_key(self):
        with tempfile.TemporaryDirectory() as dir_name:
            key_path = Path(dir_name, "private.key")
            jwt_gen_key(key_path)
            self.assertEqual(stat.filemode(key_path.stat().st_mode), "-r--------")
            self.assertEqual(key_path.stat().st_size, 64)

    def test_gen_key_permission_error(self):
        if os.geteuid() == 0:
            self.skipTest("Cannot test permission error as root")
        with tempfile.TemporaryDirectory() as dir_name:
            key_path = Path(dir_name, "private.key")
            os.chmod(dir_name, 0o000)
            with self.assertRaisesRegex(
                JWTPrivateKeyGeneratorError,
                r"^Error while generating JWT key .+/private.key: \[Errno 13\] "
                r"Permission denied: '.+/private.key'$",
            ):
                jwt_gen_key(key_path)
            # Restore mode on temporary so it can be removed without permissions
            # error.
            os.chmod(dir_name, 0o755)

    def test_gen_key_parent_file(self):
        key_path = Path("/dev/null/fail.key")
        with self.assertRaisesRegex(
            JWTPrivateKeyGeneratorError,
            r"^Error while generating JWT key /dev/null/fail.key: \[Errno 20\] Not "
            r"a directory: '/dev/null/fail.key'$",
        ):
            jwt_gen_key(key_path)

    def test_gen_key_parent_not_found(self):
        key_path = Path("/dev/fail/fail.key")
        with self.assertRaisesRegex(
            JWTPrivateKeyGeneratorError,
            r"^Error while generating JWT key /dev/fail/fail.key: \[Errno 2\] No such "
            r"file or directory: '/dev/fail/fail.key'$",
        ):
            jwt_gen_key(key_path)


class TestJWTValidateExpiration(unittest.TestCase):
    def test_ok(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTBaseManager("HS256", loader)
        token = manager.generate(1, {"user": "test"})
        payload = jwt_validate_expiration(token)
        self.assertCountEqual(payload.keys(), ["iat", "exp", "user"])
        self.assertEqual(payload["user"], "test")

    def test_token_expired(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTBaseManager("HS256", loader)
        token = manager.generate(-1, {"user": "test"})
        with self.assertRaisesRegex(
            JWTDecodeError,
            r"^Token is expired$",
        ):
            jwt_validate_expiration(token)

    def test_token_invalid(self):
        with self.assertRaisesRegex(
            JWTDecodeError,
            r"^Unable to decode token: Not enough segments$",
        ):
            jwt_validate_expiration("fail")


class TestJWTPrivateKeyFileLoader(unittest.TestCase):
    def test_load_value(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        self.assertEqual(loader.key, PRIVATE_KEY)
        self.assertEqual(loader.path, None)

    def test_load_noarg(self):
        with self.assertRaisesRegex(
            JWTPrivateKeyLoaderError,
            r"Either key value or a path must be given to load JWT private key",
        ):
            JWTPrivateKeyFileLoader()

    def test_load_path(self):
        with tempfile.NamedTemporaryFile() as fh:
            fh.write(b"SECR3T")
            fh.flush()
            loader = JWTPrivateKeyFileLoader(path=Path(fh.name))
        self.assertEqual(loader.key, b"SECR3T")

    def test_load_path_text(self):
        with tempfile.NamedTemporaryFile() as fh:
            fh.write(b"SECR3T")
            fh.flush()
            loader = JWTPrivateKeyFileLoader(path=Path(fh.name), text=True)
        self.assertEqual(loader.key, "SECR3T")

    def test_load_empty_key(self):
        with self.assertRaisesRegex(
            JWTPrivateKeyLoaderError,
            r"Key loaded from file /dev/null is empty",
        ):
            JWTPrivateKeyFileLoader(path=Path("/dev/null"))

    def test_load_unexisting_path(self):
        with self.assertRaisesRegex(
            JWTPrivateKeyLoaderError, r"Token private key file /dev/not-found not found"
        ):
            JWTPrivateKeyFileLoader(path=Path("/dev/not-found"))

    def test_load_path_permission_denied(self):
        if os.geteuid() == 0:
            self.skipTest("Cannot test permission error as root")
        with tempfile.NamedTemporaryFile() as fh:
            os.chmod(fh.name, 0o000)
            with self.assertRaisesRegex(
                JWTPrivateKeyLoaderError,
                rf"Permission error to access private key file {fh.name}",
            ):
                JWTPrivateKeyFileLoader(path=Path(fh.name))

    def test_load_path_unicode_binary_ok(self):
        with tempfile.NamedTemporaryFile() as fh:
            fh.write(b"\x12\x34\x56\x78\x9a")
            fh.flush()
            JWTPrivateKeyFileLoader(path=Path(fh.name))

    def test_load_path_unicode_test_error(self):
        with tempfile.NamedTemporaryFile() as fh:
            fh.write(b"\x12\x34\x56\x78\x9a")
            fh.flush()
            with self.assertRaisesRegex(
                JWTPrivateKeyLoaderError,
                rf"Unable to decode private key file {fh.name}: '\S+' codec can't "
                r"decode byte 0x9a in position 4: invalid start byte$",
            ):
                JWTPrivateKeyFileLoader(path=Path(fh.name), text=True)

    def test_load_create(self):
        with tempfile.TemporaryDirectory() as dir_name:
            key_path = Path(dir_name, "private.key")
            loader = JWTPrivateKeyFileLoader(path=key_path, create=True)
            self.assertEqual(stat.filemode(key_path.stat().st_mode), "-r--------")
        self.assertEqual(len(loader.key), 64)
        self.assertEqual(str(loader.path), str(key_path))

    def test_load_create_parent(self):
        with tempfile.TemporaryDirectory() as dir_name:
            key_path = Path(dir_name, "private.key")
            os.rmdir(dir_name)
            loader = JWTPrivateKeyFileLoader(
                path=key_path, create=True, create_parent=True
            )
            self.assertEqual(stat.filemode(key_path.stat().st_mode), "-r--------")
        self.assertEqual(len(loader.key), 64)
        self.assertEqual(str(loader.path), str(key_path))

    def test_load_missing_parent(self):
        # Get random temporary filename without creating it.
        dir_name = next(tempfile._get_candidate_names())
        key_path = Path(dir_name, "private.key")
        with self.assertRaisesRegex(
            JWTPrivateKeyLoaderError,
            rf"Token private key parent directory {dir_name} not found",
        ):
            JWTPrivateKeyFileLoader(path=key_path, create=True)

    def test_load_create_parent_permission_denied(self):
        if os.geteuid() == 0:
            self.skipTest("Cannot test permission error as root")
        with tempfile.TemporaryDirectory() as dir_name:
            key_path = Path(dir_name, "subdir", "private.key")
            key_path.parent.parent.chmod(0o500)
            with self.assertRaisesRegex(
                JWTPrivateKeyLoaderError,
                rf"Permission denied to create key parent directory {key_path.parent}",
            ):
                JWTPrivateKeyFileLoader(path=key_path, create=True, create_parent=True)


class TestJWTBaseManager(unittest.TestCase):
    def test_generate_decode(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTBaseManager("HS256", loader)
        token = manager.generate(1)
        self.assertCountEqual(manager.decode(token).keys(), ["iat", "exp"])

    def test_generate_decode_claimset(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTBaseManager("HS256", loader)
        token = manager.generate(1, {"user": "test"})
        payload = manager.decode(token)
        self.assertCountEqual(payload.keys(), ["iat", "exp", "user"])
        self.assertEqual(payload["user"], "test")

    def test_invalid_algo(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTBaseManager("FAIL", loader)
        with self.assertRaisesRegex(
            JWTEncodeError,
            r"^JWT token encode error: Algorithm not supported$",
        ):
            manager.generate(1)

    def test_invalid_token(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTBaseManager("HS256", loader)
        with self.assertRaisesRegex(
            JWTDecodeError,
            r"^Unable to decode token: Not enough segments$",
        ):
            manager.decode("FAIL")

    def test_invalid_signature(self):
        loader1 = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager1 = JWTBaseManager("HS256", loader1)
        loader2 = JWTPrivateKeyFileLoader(value=b"OTHER_PRIVATE_KEY")
        manager2 = JWTBaseManager("HS256", loader2)
        token = manager1.generate(1)
        with self.assertRaisesRegex(
            JWTDecodeError,
            r"^Token signature is invalid$",
        ):
            manager2.decode(token)


class TestJWTManager(unittest.TestCase):
    def test_generate_decode(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTManager("test", "HS256", loader)
        token = manager.generate(AuthenticatedUser(login="user", groups=["group"]), 1)
        self.assertEqual(manager.decode(token).login, "user")

    def test_generate_decode_anonymous(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTManager("test", "HS256", loader)
        token = manager.generate(AnonymousUser(), 1)
        self.assertEqual(manager.decode(token).login, AnonymousUser.LOGIN)

    def test_invalid_algo(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTManager("test", "FAIL", loader)
        with self.assertRaisesRegex(
            JWTEncodeError,
            r"^JWT token encode error: Algorithm not supported$",
        ):
            manager.generate(AuthenticatedUser(login="user", groups=["group"]), 1)

    def test_invalid_token(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager = JWTManager("test", "HS256", loader)
        with self.assertRaisesRegex(
            JWTDecodeError,
            r"^Unable to decode token: Not enough segments$",
        ):
            manager.decode("FAIL")

    def test_invalid_signature(self):
        loader1 = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager1 = JWTManager("test", "HS256", loader1)
        loader2 = JWTPrivateKeyFileLoader(value=b"OTHER_PRIVATE_KEY")
        manager2 = JWTManager("test", "HS256", loader2)
        token = manager1.generate(AuthenticatedUser(login="user", groups=["group"]), 1)
        with self.assertRaisesRegex(
            JWTDecodeError,
            r"^Token signature is invalid$",
        ):
            manager2.decode(token)

    def test_invalid_audience(self):
        loader = JWTPrivateKeyFileLoader(value=PRIVATE_KEY)
        manager1 = JWTManager("test1", "HS256", loader)
        manager2 = JWTManager("test2", "HS256", loader)
        token = manager1.generate(AuthenticatedUser(login="user", groups=["group"]), 1)
        with self.assertRaisesRegex(
            JWTDecodeError,
            r"^Token audience is invalid$",
        ):
            manager2.decode(token)



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from rfl.core.errors import RFLError


class SettingsError(RFLError):
    pass


class SettingsDefinitionError(SettingsError):
    pass


class SettingsSiteLoaderError(SettingsError):
    pass


class SettingsOverrideError(SettingsError):
    pass
